//
//  $Id: BinaryTree.m,v 1.4 2007/04/02 04:56:05 rh Exp $
//  A simple Binary Tree class implementation
//
//  Copyright 2006 Rene Hexel. All rights reserved.
//

#import "BinaryTree.h"
#import "BinaryTreeNode.h"
#import "SimpleStack.h"
#import "SimpleQueue.h"
#include <stdarg.h>

/*
 * the actual tree class
 */
@implementation BinaryTree

- initWithObjects: (id) obj, ...
{
	NSMutableArray *array = [NSMutableArray new];

	if (array == nil)
	{
		[self release];
		return nil;
	}

	va_list ap;                             // variable argument pointer
        va_start(ap, obj);
        while (obj)                             // for each passed argument
        {
		[array addObject: obj];		// add to array
                obj = va_arg(ap, id);           // next argument
        }
        va_end(ap);

	return [self initWithArray: [array autorelease]];
}


- initWithArray: (NSArray *) array
{
	return [self initWithEnumerator: [array objectEnumerator]
				  count: [array count]];
}


-initWithEnumerator: (NSEnumerator *) e count: (int) n
{
	if (!(self = [self init])) return nil;

	_BinaryTreeNode *rootNode = [_BinaryTreeNode treeWithEnumerator: e
								  count: n];
	if (!rootNode)
	{
		[self release];
		return nil;
	}

	[self setRoot: rootNode];
	
	return self;
}


- (NSEnumerator *) objectEnumerator
{
	return [[[defaultEnumerator alloc] initWithTree: self] autorelease];
}


- (BinaryTreePreorderEnumerator *) preorderEnumerator
{
	return [[[BinaryTreePreorderEnumerator alloc] initWithTree: self]
		autorelease];
}


- (BinaryTreePostorderEnumerator *) postorderEnumerator
{
	return [[[BinaryTreePostorderEnumerator alloc] initWithTree: self]
		autorelease];
}


- (BinaryTreeInorderEnumerator *) inorderEnumerator
{
	return [[[BinaryTreeInorderEnumerator alloc] initWithTree: self]
		autorelease];
}


- (BinaryTreeBreadthFirstEnumerator *) breadthFirstEnumerator
{
	return [[[BinaryTreeBreadthFirstEnumerator alloc] initWithTree: self]
		autorelease];
}


@end


/*
 * An object enumerator for the above tree class
 */

@implementation BinaryTreeEnumerator

- (void) dealloc
{
	[coll release];
	[super dealloc];
}


- setFirstChild: object
{
	return [self setLeftChild: object];
}


- setLastChild: object
{
	return [self setRightChild: object];
}


- setLeftChild: object
{
	_BinaryTreeNode *node = [_BinaryTreeNode new];
	[node setObject: object];
	[node setParentNode: currentNode];
	[node setLeftChild: [currentNode leftChild]];
	[currentNode setLeftChild: node];

	return self;
}


- setRightChild: object
{
	_BinaryTreeNode *node = [_BinaryTreeNode new];
	[node setObject: object];
	[node setParentNode: currentNode];
	[node setRightChild: [currentNode rightChild]];
	[currentNode setRightChild: node];

	return self;
}

@end



@implementation BinaryTreePreorderEnumerator

- init
{
	if (!(self = [super init]))
		return nil;
	
	if (!(coll = [SimpleStack new]))
	{
		[self release];
		return nil;
	}
	
	return self;
}


- advance
{
	if (currentNode == nil)			// push the topmost node
		[coll pushObject: nextNode];

	currentNode = nextNode;
	nextNode = [nextNode leftChild];	// go down left if possible

	if (nextNode)				// was there a left child?
	{
		[coll pushObject: nextNode];
		return self;
	}

	/*
	 * we have exhausted our left children, so we need to go up the coll
	 * until it is possible to go right once
	 */
	while (!nextNode && ![coll isEmpty])
		nextNode = [[coll popObject] rightChild];

	if (nextNode)				// was there a right child?
	{
		[coll pushObject: nextNode];
		return self;
	}

	return nil;				// we have run out of nodes
}

@end


@implementation BinaryTreeInorderEnumerator

- init
{
	if (!(self = [super init]))
		return nil;
	
	if (!(coll = [SimpleStack new]))
	{
		[self release];
		return nil;
	}
	
	return self;
}


- advance
{
	while (nextNode)
	{
		[coll pushObject: nextNode];	// pushing every node on the way
		nextNode = [nextNode leftChild];// down left as far as possible
	}

	if (![coll isEmpty])			// are there nodes left?
	{
		currentNode = [coll popObject];
		nextNode = [currentNode rightChild];
		return self;
	}

	currentNode = nil;

	return nil;				// we have run out of nodes
}

@end


@implementation BinaryTreePostorderEnumerator

- init
{
	if (!(self = [super init]))
		return nil;
	
	if (!(coll = [SimpleStack new]))
	{
		[self release];
		return nil;
	}
	
	return self;
}


enum action { LEFT, RIGHT, VISIT };

- advance
{
	if (currentNode == nil)			// just starting, push top node
	{
		[coll pushObject: nextNode];
		[coll pushObject: [NSNumber numberWithInt: LEFT]];
	}

	while (![coll isEmpty])		// are there nodes left?
	{
		int action = [[coll popObject] intValue];
		nextNode = [coll popObject];

		switch(action)
		{
		    case LEFT:			// push node and left subtree
			    [coll pushObject: nextNode];
			    [coll pushObject: [NSNumber numberWithInt: RIGHT]];
			    if (![nextNode leftChild]) break;
			    [coll pushObject: [nextNode leftChild]];
			    [coll pushObject: [NSNumber numberWithInt: LEFT]];
			    break;

		    case RIGHT:			// push node and right subtree
			    [coll pushObject: nextNode];
			    [coll pushObject: [NSNumber numberWithInt: VISIT]];
			    if (![nextNode rightChild]) break;
			    [coll pushObject: [nextNode rightChild]];
			    [coll pushObject: [NSNumber numberWithInt: LEFT]];
			    break;

		    case VISIT:
			    currentNode = nextNode;
			    return self;
		}
	}

	currentNode = nil;

	return nil;				// we have run out of nodes
}

@end


@implementation BinaryTreeBreadthFirstEnumerator

- init
{
	if (!(self = [super init]))
		return nil;
	
	if (!(coll = [SimpleQueue new]))	// breadth-first needs a queue
	{
		[self release];
		return nil;
	}
	
	return self;
}


- advance
{
	if (currentNode == nil)			// just starting, push top node
		[coll pushObject: nextNode];

	if ([coll isEmpty])			// are we finished?
		nextNode = nil;
	else
	{
		nextNode = [coll popObject];	// get the next node
		id node = [nextNode leftChild];
		if (node) [coll pushObject: node];
		node = [nextNode rightChild];
		if (node) [coll pushObject: node];
	}

	currentNode = nextNode;

	return self;
}

@end


